/******************************************************************************

  PS/2 Keyboard/Mouseprotokoll, MEGA8, 11.0592MHz, 
  
  PS/2 Takt: 16.6kHz
  -> 30us Low Pegel, 30us High Pegel

******************************************************************************/
#include <avr/io.h>
#include <inttypes.h>
#include <avr/delay.h>
#include "ps2.h"

//#define  __TEST


#ifdef __TEST
#include <avr/interrupt.h>
#include "strio.h"
#include "usart.h"
#endif // TEST



#define   PORT		  PORTD
#define   DDR		    DDRD
#define   PIN		    PIND

#define   DATABIT	  (PD4)		
#define   CLOCKBIT	(PD2)



static inline void SetClockBitLow(void)        
{
  cbi (PORT, CLOCKBIT);   // Clock Low (Pullup aus)
  sbi (DDR, CLOCKBIT);    // Clock Ausgang  
}

static inline void SetClockBitTristate(void)    
{
  cbi (DDR, CLOCKBIT);    // Clock Eingang
  sbi (PORT, CLOCKBIT);   // Clock Pullup einschalten
}

static inline void SetDataBitLow(void)        
{
  cbi (PORT, DATABIT);    // Data Low (Pullup aus)
  sbi (DDR, DATABIT);     // Data Ausgang
}

static inline void SetDataBitTristate(void)      
{
  cbi (DDR, DATABIT);     // Data Eingang
  sbi (PORT, DATABIT);    // Data Pullup einschalten
}


// Delays fuer 11.0592 MHz; 1Clk = 90.4ns
// 5us = 55 Takte
static void Delay5us(void)
{
  _delay_loop_1(18); 
}

// 25us = 276 Takte
static void Delay25us(void)
{
  _delay_loop_1(92); 

}

// 30us = 332 Takte
static void Delay30us(void)
{
  _delay_loop_1(110); 
}



/*
 *  sendet einen '0' oder '1' auf PS/2, je nach Parameter
 */
static void SendBit(uint8_t bit)
{
  if (!bit) {
    SetDataBitLow();
  }
  Delay5us();  
  SetClockBitLow();  
  Delay30us();
  SetClockBitTristate();
  Delay25us();  
  SetDataBitTristate();
}

/*
 *  liesst ein einzelnes Bit von der PS/2
 */
static uint8_t GetBit(void)
{
  uint8_t result; 

  Delay5us();  
  result = (PIN & _BV(DATABIT)) ? 1 : 0;  
  Delay25us();  
  SetClockBitLow();
  Delay30us();    
  SetClockBitTristate();
  return result;
}


/*
 *  ein Host Request besteht aus 100us Clock Low, anschliessend wird
 *  die Clock High, Data Low. Client muss innerhalb von 10ms Daten 
 *  abnehmen.
 * 1, wenn Request ansteht, sonst 0
 */
uint8_t CheckForHostRequest(void)
{
  SetClockBitTristate();
  SetDataBitTristate();
  if (!(PIN & _BV(DATABIT ))) {
    // Data = Low
    return 1;
  } else {
    return 0;
  }  
}


// Sende ein einzelnes Byte auf der PS/2 Leitung
void SendPS2Byte(uint8_t Data)
{
  uint8_t BitCount = 8;
  uint8_t ParityCount = 0;


	sputs ("S: ");
	printHEX(Data);
	sputs ("\n");

	// Auf Clock = High warten
	loop_until_bit_is_set(PIN, CLOCKBIT);
	
  SendBit(0);               // Startbit  
	
  while (BitCount--) {      // Datenbits
    if (Data & 0x01) {
      ParityCount++;
      SendBit(1);
    } else {
      SendBit(0);     
    }         
    Data >>= 1;
  }
  SendBit(!(ParityCount & 0x01)); // Parity
//  SetDataBitTristate();     // Clock fuer Stopbit vom HOST
  SendBit(1);               // Stop. Clock und Data auf Tristate
}


// Lese ein einzelnes Byte auf der PS/2 Leitung
// Return:  1 - OK
//          0 - Fehler (Start/Stop/Parity) 
uint8_t ReadPS2Byte(uint8_t *Data)
{
  uint8_t BitCount = 8;
  uint8_t ParityCount = 0;
  uint8_t Received = 0;
	uint8_t result = 0;

  if (GetBit()) return 1;     // Kein Startbit erkannt
  while (BitCount--) {        // Daten lesen 
    Received >>= 1;
    if (GetBit()) {
      Received |= 0x80;
      ParityCount++;
    }
  }
	*Data = Received;
  if (!((GetBit() + ParityCount) & 0x01)) {
		result = 1;  // Parity Error
	}
  if (!GetBit()) {
		result = 1;    // Stopbit fehlt
	}
  SendBit(0);                 // Ack senden
  SetDataBitTristate();

	sputs ("R: ");
	printHEX(result);
	sputs ("\n");


  return result;
}


void initPS2IO(void)
{
  // Pullup enable
  SFIOR &= ~_BV(PUD);  
  // Clock & Data high impedance
  DDR &= ~(_BV(DATABIT) | _BV(CLOCKBIT));
  // Clock & Data Low -> preparation for open collector output
  PORT |= (_BV(DATABIT) | _BV(CLOCKBIT));
}

#ifdef __TEST

int main (int argc, char **argv)
{
  uint8_t Data;
	uint16_t i;

  initPS2IO();
	usart_init();
  sei();

	sputs ("Start!\n");

	_delay_loop_2(5000); // ~1800us

	sputs ("0xaa send\n");
	SendPS2Byte(0xaa);

	i = 0;
	while (1) {
		_delay_loop_2(10000); // ~3.6ms
		if (CheckForHostRequest()) {
			if (ReadPS2Byte(&Data)) {
				usart_transmit('E');
				printHEX(Data);
				usart_transmit('\n');
				_delay_loop_2(5000); // ~1800us
				SendPS2Byte(0xFe);
			} else {
				usart_transmit('O');
				printHEX(Data);
				usart_transmit('\n');
				_delay_loop_2(5000); // ~1800us
				SendPS2Byte(0xFa);
			}    
		} 
	}


  return 0;
}

#endif // __TEST


