/****************************************************************************
*
* Timerhandling. 
* Zweck:  interner TickCounter 
*         Sekunden Counter fr Wakeup etc.
*
* 11.0692MHz, Timer 0 mit 1024 Vorteiler
* -> d.h. Taktung Timer0 mit 10.8KHz -> TCNT0 muss 92us (1024 Takte) nach 
*    Interrupt neu geladen sein.
* dann Teilung durch 216    
* -> Overflow alle 50Hz = 20ms Periode
*
* Wenn der Code getestet wurde sollte das ganze in Assembler abgebildet werden. 
*
****************************************************************************/

// #define TEST


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <inttypes.h>


#ifdef TEST
#include "usart.h"
#include "strio.h"
#endif


#define OVERFLOWVALUE (216)

/* Interrupt handlers */

static int32_t tickcount = 0;
static int32_t alertcounter = 0;
static void  (* alertfunction)(void);  

/*
 * Liefert die Diffenz zwischen aktuellem Timestamp und 
 * uebegebenen.
 */
int16_t GetTickDiff(int32_t oldtick)
{
  int16_t result;
  cli();
  result = (int16_t) tickcount - oldtick;
  sei();
  return result;
}


/*
 *  Setzen einer Alarmfunktion, angabe des Auslseticks und der Funktion,
 *  die bei Faelligkeit aufgerufen wird.
 */
void SetAlert(int32_t alerttime, void  (* alertfkt)(void))
{
  alertfunction = alertfkt;
  alertcounter = alerttime;
}

/*
 *  Abfrage des aktuellen Tickcounts als 16 Bit Wert
 */
int16_t GetShortTicks(void)
{
  int16_t result;
  cli();
  result = (int16_t)tickcount;
  sei();
  return result;
}


/*
 *  Abfrage des aktuellen Tickcounts als 32 Bit Wert
 */
int32_t GetTicks(void)
{
  int32_t result;
  cli();
  result = tickcount;
  sei();
  return result;
}


SIGNAL(SIG_OVERFLOW0)
{
  TCNT0 = (256 - OVERFLOWVALUE);  // Einstellung des Wertes fuer Overflow
  tickcount++;
  if (alertcounter) {
    if (tickcount >= alertcounter) {
      alertcounter = 0;
      if (alertfunction) {
        alertfunction();
      }
    }
  }
}

/*
 *  Init Timer0 mit Vorteiler 
 */
void init_timer0(void)
{
  TCNT0 = (256 - OVERFLOWVALUE);  // Einstellung des Wertes fuer Overflow
  TCCR0 = 5;                      // Vorteiler auf 1024
  TIMSK |= _BV(TOIE0);            // Overflow Interrupt enablen 
                                  // sei() Global notwendig!  
}


#ifdef TEST

void callback (void)
{
	puts ("Callback!\n\n");
}


int main (int argc, char **argv)
{
	int32_t tmptick;

	init_timer0();
	usart_init ();

	sei();

	puts ("Start!\n\n");

	tmptick = GetTicks();
	tmptick += 2*50;		// 1 Min warten

	SetAlert(tmptick, callback);

	while (1);

	return 0;
}

#endif // TEST

